from setuptools import Extension, Distribution
import numpy
from Cython.Build import cythonize
from Cython.Distutils.build_ext import new_build_ext as cython_build_ext


def build():
    extensions = [
        Extension(
            'pyobs_sbig.sbigudrv',
            ['pyobs_sbig/sbigudrv.pyx', 'src/csbigcam.cpp', 'src/csbigimg.cpp'],
            libraries=['sbigudrv', 'cfitsio'],
            include_dirs=[numpy.get_include(), '/usr/include/cfitsio'],
            extra_compile_args=['-fPIC']
        )
    ]
    ext_modules = cythonize(
        extensions,
        compiler_directives={
            "binding": True,
            "language_level": "3"
        },
    )

    distribution = Distribution({
        "name": "extended",
        "ext_modules": ext_modules,
        "cmdclass": {
            "build_ext": cython_build_ext,
        },
    })
    #distribution.package_dir = "extended"

    distribution.run_command("build_ext")
    build_ext_cmd = distribution.get_command_obj("build_ext")
    build_ext_cmd.copy_extensions_to_source()


if __name__ == "__main__":
    build()



import asyncio
import logging
import math
from datetime import datetime
from typing import Union, Any, Optional, Dict

from pyobs.images import Image
from pyobs.interfaces import ICamera, IWindow
from pyobs.modules.camera.basecamera import BaseCamera
from pyobs.utils.enums import ExposureStatus
from pyobs_sbig.sbigdriver import SbigDriver
from pyobs_sbig.sbigudrv import *


log = logging.getLogger(__name__)


class SbigBaseCamera(BaseCamera, ICamera, IWindow):
    """A pyobs module for SBIG cameras."""
    __module__ = 'pyobs_sbig'

    def __init__(self, sensor: Union[str, ActiveSensor] = ActiveSensor.IMAGING, driver: Optional[SbigDriver] = None,
                 driver_kwargs: Optional[Dict[str, Any]] = None, **kwargs: Any):
        """Initializes a new SbigCamera.

        Args:
            sensor: Sensor to use, if camera has more than one.
            driver_kwargs: kwargs for driver.

        """
        BaseCamera.__init__(self, **kwargs)

        # create driver?
        if driver is not None:
            self._driver = driver
        else:
            driver_kwargs = {} if driver_kwargs is None else driver_kwargs
            self._driver = self.add_child_object({'class': 'pyobs_sbig.SbigDriver'},
                                                 object_class=SbigDriver, **driver_kwargs)

        # active sensor
        if isinstance(sensor, str):
            self._active_sensor = ActiveSensor[sensor.upper()]
        elif isinstance(sensor, ActiveSensor):
            self._active_sensor = sensor
        else:
            raise ValueError('Invalid sensor given.')

        # create image
        self._img = SBIGImg()

        # window and binning
        self._full_frame = (0, 0, 0, 0)
        self._window = (0, 0, 0, 0)
        self._binning = (1, 1)

    async def open(self) -> None:
        """Open module.

        Raises:
            ValueError: If cannot connect to camera or set filter wheel.
        """
        await BaseCamera.open(self)

        # get window
        self._window = await self.get_full_frame()

    async def get_full_frame(self, **kwargs: Any) -> Tuple[int, int, int, int]:
        """Returns full size of CCD.

        Returns:
            Tuple with left, top, width, and height set.
        """
        return self._driver.full_frame(self._active_sensor)

    async def get_window(self, **kwargs: Any) -> Tuple[int, int, int, int]:
        """Returns the camera window.

        Returns:
            Tuple with left, top, width, and height set.
        """
        return self._window

    async def set_window(self, left: int, top: int, width: int, height: int, **kwargs: Any) -> None:
        """Set the camera window.

        Args:
            left: X offset of window.
            top: Y offset of window.
            width: Width of window.
            height: Height of window.
        """
        self._window = (left, top, width, height)
        log.info('Setting window to %dx%d at %d,%d...', width, height, left, top)

    async def _expose(self, exposure_time: float, open_shutter: bool, abort_event: asyncio.Event) -> Image:
        """Actually do the exposure, should be implemented by derived classes.

        Args:
            exposure_time: The requested exposure time in ms.
            open_shutter: Whether or not to open the shutter.
            abort_event: Event that gets triggered when exposure should be aborted.

        Returns:
            The actual image.

        Raises:
            ValueError: If exposure was not successful.
        """

        #  binning
        binning = self._binning

        # set window, CSBIGCam expects left/top also in binned coordinates, so divide by binning
        left = int(math.floor(self._window[0]) / binning[0])
        top = int(math.floor(self._window[1]) / binning[1])
        width = int(math.floor(self._window[2]) / binning[0])
        height = int(math.floor(self._window[3]) / binning[1])
        log.info("Set window to %dx%d (binned %dx%d) at %d,%d.",
                 self._window[2], self._window[3], width, height, left, top)
        window = (left, top, width, height)

        # get date obs
        log.info('Starting exposure with for %.2f seconds...', exposure_time)
        date_obs = datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f")

        # init image
        self._img.image_can_close = False

        # start exposure (can raise ValueError)
        self._driver.start_exposure(self._active_sensor, self._img, open_shutter, exposure_time, window=window,
                                    binning=binning)

        # wait for it
        while not self._driver.has_exposure_finished(self._active_sensor):
            # was aborted?
            if abort_event.is_set():
                raise ValueError('Exposure aborted.')
            await asyncio.sleep(0.01)

        # finish exposure
        self._driver.end_exposure(self._active_sensor)

        # wait for readout
        log.info('Exposure finished, reading out...')
        await self._change_exposure_status(ExposureStatus.READOUT)

        # start readout (can raise ValueError)
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self._driver.readout, self._active_sensor, self._img, open_shutter)

        # finalize image
        self._img.image_can_close = True

        # download data
        data = self._img.data

        # temp & cooling
        _, temp, setpoint, _ = self._driver.camera.get_cooling()

        # create FITS image and set header
        img = Image(data)
        img.header['DATE-OBS'] = (date_obs, 'Date and time of start of exposure')
        img.header['EXPTIME'] = (exposure_time, 'Exposure time [s]')
        img.header['DET-TEMP'] = (temp, 'CCD temperature [C]')
        img.header['DET-TSET'] = (setpoint, 'Cooler setpoint [C]')

        # binning
        img.header['XBINNING'] = img.header['DET-BIN1'] = (self._binning[0], 'Binning factor used on X axis')
        img.header['YBINNING'] = img.header['DET-BIN2'] = (self._binning[1], 'Binning factor used on Y axis')

        # window
        img.header['XORGSUBF'] = (self._window[0], 'Subframe origin on X axis')
        img.header['YORGSUBF'] = (self._window[1], 'Subframe origin on Y axis')

        # statistics
        img.header['DATAMIN'] = (float(np.min(data)), 'Minimum data value')
        img.header['DATAMAX'] = (float(np.max(data)), 'Maximum data value')
        img.header['DATAMEAN'] = (float(np.mean(data)), 'Mean data value')

        # biassec/trimsec
        frame = await self.get_full_frame()
        self.set_biassec_trimsec(img.header, *frame)

        # return FITS image
        log.info('Readout finished.')
        return img

    async def _abort_exposure(self) -> None:
        """Abort the running exposure. Should be implemented by derived class.

        Raises:
            ValueError: If an error occured.
        """
        await self._change_exposure_status(ExposureStatus.IDLE)


__all__ = ['SbigBaseCamera']



import logging
from typing import Union, Any

from .sbigbasecamera import SbigBaseCamera
from .sbigudrv import *


log = logging.getLogger(__name__)


class SbigTracker(SbigBaseCamera):
    """A pyobs module for the tracker chip in some SBIG cameras."""
    __module__ = 'pyobs_sbig'

    def __init__(self, sensor: Union[str, ActiveSensor] = ActiveSensor.TRACKING, **kwargs: Any):
        """Initializes a new SbigTracker.

        Args:
            setpoint: Cooling temperature setpoint.
        """
        SbigBaseCamera.__init__(self, sensor=sensor, **kwargs)


__all__ = ['SbigTracker']



import logging
from typing import Any, Dict

from pyobs.interfaces import IBinning, ICooling
from .sbigbasecamera import SbigBaseCamera
from .sbigudrv import *

log = logging.getLogger(__name__)


class SbigCamera(SbigBaseCamera, IBinning, ICooling):
    """A pyobs module for SBIG cameras."""
    __module__ = 'pyobs_sbig'

    def __init__(self, setpoint: float = -20, **kwargs: Any):
        """Initializes a new SbigCamera.

        Args:
            setpoint: Cooling temperature setpoint.
        """
        SbigBaseCamera.__init__(self, **kwargs)

        # cooling
        self._setpoint = setpoint
        self._cooling = None

    async def open(self) -> None:
        """Open module.

        Raises:
            ValueError: If cannot connect to camera or set filter wheel.
        """
        await SbigBaseCamera.open(self)

        # cooling
        await self.set_cooling(self._setpoint is not None, self._setpoint)

    async def get_binning(self, **kwargs: Any) -> Tuple[int, int]:
        """Returns the camera binning.

        Returns:
            Dictionary with x and y.
        """
        return self._binning

    async def set_binning(self, x: int, y: int, **kwargs: Any) -> None:
        """Set the camera binning.

        Args:
            x: X binning.
            y: Y binning.
        """
        self._binning = (x, y)
        log.info('Setting binning to %dx%d...', x, y)

    async def set_cooling(self, enabled: bool, setpoint: float, **kwargs: Any) -> None:
        """Enables/disables cooling and sets setpoint.

        Args:
            enabled: Enable or disable cooling.
            setpoint: Setpoint in celsius for the cooling.

        Raises:
            ValueError: If cooling could not be set.
        """

        # log
        if enabled:
            log.info('Enabling cooling with a setpoint of %.2f°C...', setpoint)
        else:
            log.info('Disabling cooling and setting setpoint to 20°C...')

        # do it
        self._driver.camera.set_cooling(enabled, setpoint)

    async def get_cooling_status(self, **kwargs: Any) -> Tuple[bool, float, float]:
        """Returns the current status for the cooling.

        Returns:
            Tuple containing:
                Enabled (bool):         Whether the cooling is enabled
                SetPoint (float):       Setpoint for the cooling in celsius.
                Power (float):          Current cooling power in percent or None.
        """

        try:
            enabled, temp, setpoint, power = self._driver.camera.get_cooling()
            self._cooling = enabled is True, setpoint, power * 100.
        except ValueError:
            # use existing cooling
            pass
        return self._cooling

    async def get_temperatures(self, **kwargs: Any) -> Dict[str, float]:
        """Returns all temperatures measured by this module.

        Returns:
            Dict containing temperatures.
        """

        try:
            _, temp, _, _ = self._driver.camera.get_cooling()
            return {'CCD': temp}
        except ValueError:
            # use existing temps
            pass


__all__ = ['SbigCamera']



# distutils: language = c++
import threading
import time
from enum import Enum
from contextlib import contextmanager
from typing import Tuple

import numpy as np
cimport numpy as np
np.import_array()

from .sbigudrv cimport *


@contextmanager
def acquire_lock(lock):
    if not lock.acquire(timeout=1):
        raise ValueError('Could not acquire exclusive lock on camera.')
    yield
    lock.release()


class FilterWheelModel(Enum):
    UNKNOWN = CFW_MODEL_SELECT.CFWSEL_UNKNOWN
    CFW2 = CFW_MODEL_SELECT.CFWSEL_CFW2
    CFW5 = CFW_MODEL_SELECT.CFWSEL_CFW5
    CFW8 = CFW_MODEL_SELECT.CFWSEL_CFW8
    CFWL = CFW_MODEL_SELECT.CFWSEL_CFWL
    CFW402 = CFW_MODEL_SELECT.CFWSEL_CFW402
    AUTO = CFW_MODEL_SELECT.CFWSEL_AUTO
    CFW6A = CFW_MODEL_SELECT.CFWSEL_CFW6A
    CFW10 = CFW_MODEL_SELECT.CFWSEL_CFW10
    CFW10_SERIAL = CFW_MODEL_SELECT.CFWSEL_CFW10_SERIAL
    CFW9 = CFW_MODEL_SELECT.CFWSEL_CFW9
    CFWL8 = CFW_MODEL_SELECT.CFWSEL_CFWL8
    CFWL8G = CFW_MODEL_SELECT.CFWSEL_CFWL8G
    CFW1603 = CFW_MODEL_SELECT.CFWSEL_CFW1603
    FW5_STX = CFW_MODEL_SELECT.CFWSEL_FW5_STX
    FW5_8300 = CFW_MODEL_SELECT.CFWSEL_FW5_8300
    FW8_8300 = CFW_MODEL_SELECT.CFWSEL_FW8_8300
    FW7_STX = CFW_MODEL_SELECT.CFWSEL_FW7_STX
    FW8_STT = CFW_MODEL_SELECT.CFWSEL_FW8_STT


class FilterWheelComPort(Enum):
    COM1 = CFW_COM_PORT.CFWPORT_COM1
    COM2 = CFW_COM_PORT.CFWPORT_COM2
    COM3 = CFW_COM_PORT.CFWPORT_COM3
    COM4 = CFW_COM_PORT.CFWPORT_COM4


class FilterWheelPosition(Enum):
    UNKNOWN = CFW_POSITION.CFWP_UNKNOWN
    CFWP_1 = CFW_POSITION.CFWP_1
    CFWP_2 = CFW_POSITION.CFWP_2
    CFWP_3 = CFW_POSITION.CFWP_3
    CFWP_4 = CFW_POSITION.CFWP_4
    CFWP_5 = CFW_POSITION.CFWP_5
    CFWP_6 = CFW_POSITION.CFWP_6
    CFWP_7 = CFW_POSITION.CFWP_7
    CFWP_8 = CFW_POSITION.CFWP_8
    CFWP_9 = CFW_POSITION.CFWP_9
    CFWP_10 = CFW_POSITION.CFWP_10


class FilterWheelStatus(Enum):
    UNKNOWN = CFW_STATUS.CFWS_UNKNOWN
    IDLE = CFW_STATUS.CFWS_IDLE
    BUSY = CFW_STATUS.CFWS_BUSY


class ActiveSensor(Enum):
    IMAGING = CCD_REQUEST.CCD_IMAGING
    TRACKING = CCD_REQUEST.CCD_TRACKING
    EXT_TRACKING = CCD_REQUEST.CCD_EXT_TRACKING


cdef class SBIGImg:
    cdef CSBIGImg* obj

    def __cinit__(self):
        self.obj = new CSBIGImg()

    @property
    def image_can_close(self) -> bool:
        return self.obj.GetImageCanClose()

    @image_can_close.setter
    def image_can_close(self, can_do: bool):
        self.obj.SetImageCanClose(can_do)

    @property
    def data(self):
        # get image dimensions
        width, height = self.obj.GetWidth(), self.obj.GetHeight()

        # create a C array to describe the shape of the ndarray
        cdef np.npy_intp shape[1]
        shape[0] = <np.npy_intp>(width * height)

        # Use the PyArray_SimpleNewFromData function from numpy to create a
        # new Python object pointing to the existing data
        arr = np.PyArray_SimpleNewFromData(1, shape, np.NPY_USHORT, <void *>self.obj.GetImagePointer())

        # reshape it to 2D
        return arr.reshape(height, width)


cdef class SBIGCam:
    cdef CSBIGCam* obj
    cdef object aborted
    cdef object lock

    def __cinit__(self):
        self.obj = new CSBIGCam(SBIG_DEVICE_TYPE.DEV_USB)

    def __init__(self):
        self.aborted = False
        self.lock = threading.Lock()

    def establish_link(self):
        res = self.obj.EstablishLink()
        if res != 0:
            raise ValueError(self.obj.GetErrorString(res))

    @property
    def sensor(self) -> ActiveSensor:
        with acquire_lock(self.lock):
            return ActiveSensor(self.obj.GetActiveCCD())

    @sensor.setter
    def sensor(self, camera: ActiveSensor):
        with acquire_lock(self.lock):
            self.obj.SetActiveCCD(camera.value)

    @property
    def readout_mode(self):
        with acquire_lock(self.lock):
            return self.obj.GetReadoutMode()

    @readout_mode.setter
    def readout_mode(self, rm):
        with acquire_lock(self.lock):
            self.obj.SetReadoutMode(rm)

    @property
    def full_frame(self) -> Tuple[int, int, int, int]:
        # define width and height
        cdef int width = 0
        cdef int height = 0

        # acquire lock
        with acquire_lock(self.lock):
            # call library
            res = self.obj.GetFullFrame(width, height)
            if res != 0:
                raise ValueError(self.obj.GetErrorString(res))
            return 0, 0, width, height

    @property
    def window(self):
        # define left, top, width and height
        cdef int left = 0
        cdef int top = 0
        cdef int width = 0
        cdef int height = 0

        # acquire lock
        with acquire_lock(self.lock):
            # call library
            self.obj.GetSubFrame(left, top, width, height)
            return left, top, width, height

    @window.setter
    def window(self, wnd):
        # set window
        with acquire_lock(self.lock):
            self.obj.SetSubFrame(wnd[0], wnd[1], wnd[2], wnd[3])

    @property
    def binning(self):
        # acquire lock
        with acquire_lock(self.lock):
            # get readout mode
            mode = self.obj.GetReadoutMode()

            # return it
            # 0 = No binning, high resolution
            # 1 = 2x2 on-chip binning, medium resolution
            # 2 = 3x3 on-chip binning, low resolution (ST-7/8/etc/237 only)
            if mode in [0, 1, 2]:
                return mode + 1, mode + 1
            else:
                raise ValueError('Unknown readout mode.')

    @binning.setter
    def binning(self, binning):
        # check
        if binning[0] != binning[1] or binning[0] < 1 or binning[0] > 3:
            raise ValueError('Only 1x1, 2x2, and 3x3 binnings supported.')

        # set it
        with acquire_lock(self.lock):
            self.obj.SetReadoutMode(binning[0] - 1)

    @property
    def exposure_time(self):
        with acquire_lock(self.lock):
            return self.obj.GetExposureTime()

    @exposure_time.setter
    def exposure_time(self, exp):
        with acquire_lock(self.lock):
            self.obj.SetExposureTime(exp)

    @property
    def temperature(self):
        cdef double temp = 0

        # acquire lock
        with acquire_lock(self.lock):
            # get temp
            res = self.obj.GetCCDTemperature(temp)
            if res != 0:
                raise ValueError(self.obj.GetErrorString(res))
            return temp

    def set_cooling(self, enable: bool, setpoint: float):
        # acquire lock
        with acquire_lock(self.lock):
            res = self.obj.SetTemperatureRegulation(enable, setpoint)
            if res != 0:
                raise ValueError(self.obj.GetErrorString(res))

    def get_cooling(self):
        # define vars
        cdef MY_LOGICAL enabled = 0
        cdef double temp = 0.
        cdef double setpoint = 0.
        cdef double power = 0.

        # acquire lock
        with acquire_lock(self.lock):
            # get it
            res = self.obj.QueryTemperatureStatus(enabled, temp, setpoint, power)
            if res != 0:
                raise ValueError(self.obj.GetErrorString(res))
            return enabled == 1, temp, setpoint, power

    def abort(self):
        # abort exposure
        self.aborted = True

    def was_aborted(self):
        return self.aborted

    def start_exposure(self, img: SBIGImg, shutter: bool):
         # define vars
        cdef MY_LOGICAL complete = 0

        # acquire lock
        with acquire_lock(self.lock):
            # not aborted
            self.aborted = False

            # get mode
            mode = SBIG_DARK_FRAME.SBDF_LIGHT_ONLY if shutter else SBIG_DARK_FRAME.SBDF_DARK_ONLY

            # do setup
            res = self.obj.GrabSetup(img.obj, mode)
            if res != 0:
                raise ValueError(self.obj.GetErrorString(res))

            # end current exposure, if any
            res = self.obj.EndExposure()
            if res != 0:
                raise ValueError(self.obj.GetErrorString(res))

            # start exposure
            shutter_cmd = SHUTTER_COMMAND.SC_OPEN_SHUTTER if shutter else  SHUTTER_COMMAND.SC_CLOSE_SHUTTER
            res = self.obj.StartExposure(shutter_cmd)
            if res != 0:
                raise ValueError('Could not start exposure: ' + self.obj.GetErrorString(res))

    def has_exposure_finished(self):
         # define vars
        cdef MY_LOGICAL complete = 0

        # acquire lock
        with acquire_lock(self.lock):
            # is complete?
            res = self.obj.IsExposureComplete(complete)
            if res != 0:
                raise ValueError('Could not get exposure status: ' + self.obj.GetErrorString(res))

            # break on error or if complete or aborted
            return res != 0 or complete

    def end_exposure(self):
        # acquire lock
        with acquire_lock(self.lock):
            res = self.obj.EndExposure()
            if res != 0:
                raise ValueError('Could not end exposure: ' + self.obj.GetErrorString(res))

    @staticmethod
    cdef int _readout(CSBIGCam *cam, CSBIGImg *img, SBIG_DARK_FRAME mode) nogil:
        cdef int res = cam.Readout(img, mode)
        return res

    def readout(self, img: SBIGImg, shutter: bool):
        # get mode
        cdef SBIG_DARK_FRAME mode = SBIG_DARK_FRAME.SBDF_LIGHT_ONLY if shutter else SBIG_DARK_FRAME.SBDF_DARK_ONLY
        cdef int res = 0

        # acquire lock
        with acquire_lock(self.lock):
            # do readout
            with nogil:
                res = int(SBIGCam._readout(self.obj, img.obj, mode))
            if res != 0:
                raise ValueError(self.obj.GetErrorString(int(res)))

    def set_filter_wheel(self, wheel: FilterWheelModel, com_port: FilterWheelComPort = FilterWheelComPort.COM1):
        # acquire lock
        with acquire_lock(self.lock):
            res = self.obj.SetCFWModel(wheel.value, com_port.value)
            if res != 0:
                raise ValueError(self.obj.GetErrorString(res))

    def set_filter(self, position: FilterWheelPosition):
        # acquire lock
        with acquire_lock(self.lock):
            res = self.obj.SetCFWPosition(position.value)
            if res != 0:
                raise ValueError(self.obj.GetErrorString(res))

    def get_filter_position_and_status(self):
        # define vars
        cdef CFW_POSITION position = CFW_POSITION.CFWP_UNKNOWN
        cdef CFW_STATUS status = CFW_STATUS.CFWS_UNKNOWN

        # acquire lock
        with acquire_lock(self.lock):
            # request from driver
            res = self.obj.GetCFWPositionAndStatus(position, status)
            if res != 0:
                raise ValueError(self.obj.GetErrorString(res))

            # parse it, since for whatever reason, position can be a value not defined in CFW_POSITION...
            try:
                pos = FilterWheelPosition(position)
            except ValueError:
                pos = FilterWheelPosition.UNKNOWN

            # return it
            return pos, FilterWheelStatus(status)



import logging
import threading
from typing import Any, Optional

from pyobs.images import Image
from .sbigfiltercamera import SbigFilterCamera


log = logging.getLogger(__name__)


class Sbig6303eCamera(SbigFilterCamera):
    """A pyobs module for SBIG6303e cameras."""
    __module__ = 'pyobs_sbig'

    async def _expose(self, exposure_time: float, open_shutter: bool, abort_event: threading.Event) -> Image:
        """Actually do the exposure, should be implemented by derived classes.

        Args:
            exposure_time: The requested exposure time in ms.
            open_shutter: Whether or not to open the shutter.
            abort_event: Event that gets triggered when exposure should be aborted.

        Returns:
            The actual image.

        Raises:
            ValueError: If exposure was not successful.
        """

        # do expsure
        img = await SbigFilterCamera._expose(self, exposure_time, open_shutter, abort_event)

        # get binning
        xbin, ybin = await self.get_binning()

        # gain is different in binned images
        gain = (1.4, 'Detector gain [e-/ADU]') if xbin == ybin == 1 else (2.3, 'Detector gain [e-/ADU]')
        img.header['DET-GAIN'] = gain

        # finished
        return img

    async def init(self, **kwargs: Any) -> None:
        pass

    async def park(self, **kwargs: Any) -> None:
        pass

    async def stop_motion(self, device: Optional[str] = None, **kwargs: Any) -> None:
        pass

    async def is_ready(self, **kwargs: Any) -> bool:
        return True


__all__ = ['Sbig6303eCamera']



"""
TODO: write doc
"""
__title__ = 'SBIG camera modules'

from .sbigbasecamera import SbigBaseCamera
from .sbigdriver import SbigDriver
from .sbigcamera import SbigCamera
from .sbigtracker import SbigTracker
from .sbigfiltercamera import SbigFilterCamera
from .sbig6303e import Sbig6303eCamera



import logging
import threading
from typing import List, Optional, Any

from pyobs.images import Image
from pyobs.mixins import MotionStatusMixin
from pyobs.events import FilterChangedEvent
from pyobs.interfaces import IFilters
from pyobs.utils.enums import MotionStatus
from pyobs.utils.threads import LockWithAbort

from .sbigcamera import SbigCamera
from .sbigudrv import FilterWheelPosition, FilterWheelStatus, FilterWheelModel


log = logging.getLogger(__name__)


class SbigFilterCamera(MotionStatusMixin, SbigCamera, IFilters):
    """A pyobs module for SBIG cameras."""
    __module__ = 'pyobs_sbig'

    def __init__(self, filter_wheel: str, filter_names: Optional[List[str]] = None, **kwargs: Any):
        """Initializes a new SbigCamera.

        Args:
            filter_names: List of filter names.
        """
        SbigCamera.__init__(self, **kwargs, driver_kwargs=dict(filter_wheel=filter_wheel))

        # and filter names
        if filter_names is None:
            filter_names = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10']
        positions = [p for p in FilterWheelPosition]
        self._filter_names = dict(zip(positions[1:], filter_names))
        self._filter_names[FilterWheelPosition.UNKNOWN] = 'UNKNOWN'

        # allow to abort motion (filter wheel)
        self._lock_motion = threading.Lock()
        self._abort_motion = threading.Event()

        # current position
        self._position = FilterWheelPosition.UNKNOWN

        # init mixins
        MotionStatusMixin.__init__(self, **kwargs, motion_status_interfaces=['IFilters'])

    async def open(self) -> None:
        """Open module.

        Raises:
            ValueError: If cannot connect to camera or set filter wheel.
        """

        # open camera
        await SbigCamera.open(self)

        # init status of filter wheel
        if self._driver.filter_wheel != FilterWheelModel.UNKNOWN:
            await self._change_motion_status(MotionStatus.POSITIONED, interface='IFilters')

        # subscribe to events
        if self.comm:
            await self.comm.register_event(FilterChangedEvent)

    async def _expose(self, exposure_time: float, open_shutter: bool, abort_event: threading.Event) -> Image:
        """Actually do the exposure, should be implemented by derived classes.

        Args:
            exposure_time: The requested exposure time in ms.
            open_shutter: Whether or not to open the shutter.
            abort_event: Event that gets triggered when exposure should be aborted.

        Returns:
            The actual image.

        Raises:
            ValueError: If exposure was not successful.
        """

        # do expsure
        img = await SbigCamera._expose(self, exposure_time, open_shutter, abort_event)

        # add filter to FITS headers
        if self._driver.filter_wheel != FilterWheelModel.UNKNOWN:
            img.header['FILTER'] = (await self.get_filter(), 'Current filter')

        # finished
        return img

    async def set_filter(self, filter_name: str, **kwargs: Any) -> None:
        """Set the current filter.

        Args:
            filter_name: Name of filter to set.

        Raises:
            ValueError: If binning could not be set.
            NotImplementedError: If camera doesn't have a filter wheel.
        """

        # do we have a filter wheel?
        if self._driver.filter_wheel == FilterWheelModel.UNKNOWN:
            raise NotImplementedError

        # reverse dict and search for name
        filters = {y: x for x, y in self._filter_names.items()}
        if filter_name not in filters:
            raise ValueError('Unknown filter: %s' % filter_name)

        # there already?
        position, status = self._driver.camera.get_filter_position_and_status()
        if position == filters[filter_name] and status == FilterWheelStatus.IDLE:
            log.info('Filter changed.')
            return

        # set status
        await self._change_motion_status(MotionStatus.SLEWING, interface='IFilters')

        # acquire lock
        with LockWithAbort(self._lock_motion, self._abort_motion):
            # set it
            log.info('Changing filter to %s...', filter_name)
            self._driver.camera.set_filter(filters[filter_name])

            # wait for it
            while True:
                # break, if wheel is idle and filter is set
                position, status = self._driver.camera.get_filter_position_and_status()
                if position == filters[filter_name] and status == FilterWheelStatus.IDLE:
                    break

                # abort?
                if self._abort_motion.is_set():
                    log.warning('Filter change aborted.')
                    return

            # send event
            log.info('Filter changed.')
            self.comm.send_event(FilterChangedEvent(filter_name))

        # set status
        await self._change_motion_status(MotionStatus.POSITIONED, interface='IFilters')

    async def get_filter(self, **kwargs: Any) -> str:
        """Get currently set filter.

        Returns:
            Name of currently set filter.

        Raises:
            ValueError: If filter could not be fetched.
            NotImplementedError: If camera doesn't have a filter wheel.
        """

        # do we have a filter wheel?
        if self._driver.filter_wheel == FilterWheelModel.UNKNOWN:
            raise NotImplementedError

        try:
            self._position, _ = self._driver.camera.get_filter_position_and_status()
        except ValueError:
            # use existing position
            pass
        return self._filter_names[self._position]

    async def list_filters(self, **kwargs: Any) -> List[str]:
        """List available filters.

        Returns:
            List of available filters.

        Raises:
            NotImplementedError: If camera doesn't have a filter wheel.
        """

        # do we have a filter wheel?
        if self._driver.filter_wheel == FilterWheelModel.UNKNOWN:
            raise NotImplementedError

        # return names
        return [f for f in self._filter_names.values() if f is not None]


__all__ = ['SbigFilterCamera']



from libcpp.string cimport string
from enum import Enum


cdef extern from "../src/csbigimg.h":
    ctypedef unsigned short MY_LOGICAL

    cdef cppclass CSBIGImg:
        CSBIGImg()
    
        void			Init()
        void			DeleteImageData()
    
        # Accessor Functions
        int				GetHeight()
        int				GetWidth()
        unsigned short	*GetImagePointer()
        #struct tm		GetImageStartTime()
        void			SetCCDTemperature(double temp)
        double			GetCCDTemperature()
        void			SetExposureTime(double exp)
        double			GetExposureTime()
        void			SetEachExposure(double exp)
        double			GetEachExposure()			
        void			SetFocalLength(double fl)	
        double			GetFocalLength()			
        void			SetApertureArea(double ap)	
        double			GetApertureArea()			
        void			SetResponseFactor(double resp)
        double			GetResponseFactor()	
        void			SetPixelHeight(double ht)
        double			GetPixelHeight()		
        void			SetPixelWidth(double wd)
        double			GetPixelWidth()	
        void			SetEGain(double gn)
        double			GetEGain()			
        void			SetBackground(long back)
        long			GetBackground()			
        void			SetRange(long range)	
        long			GetRange()				
        void			SetSaturationLevel(unsigned short sat)
        unsigned short	GetSaturationLevel()	
        void			SetNumberExposures(unsigned short no)
        unsigned short	GetNumberExposures()	
        void			SetTrackExposure(double exp)
        double			GetTrackExposure()		
        void			SetReadoutMode(unsigned short rm)
        unsigned short	GetReadoutMode()		
        void			SetPedestal(unsigned short ped)
        unsigned short	GetPedestal()			
        void			SetExposureState(unsigned short es)
        unsigned short	GetExposureState()		
        void			SetImageNote(string str)
        string			GetImageNote()			
        void			SetObserver(string str)	
        string			GetObserver()			
        void			SetHistory(string str)	
        string			GetHistory()			
        void			SetFilter(string str)	
        string			GetFilter()				
        void			SetSoftware(string str)	
        string			GetSoftware()			
        void			SetCameraModel(string str)
        string			GetCameraModel()		
        MY_LOGICAL		GetImageModified()
        void			SetImageModified(MY_LOGICAL isMod)
        #SBIG_IMAGE_FORMAT GetDefaultImageFormat()
        #void			SetDefaultImageFormat(SBIG_IMAGE_FORMAT fmt)
        MY_LOGICAL		GetImageCanClose()
        void			SetImageCanClose(MY_LOGICAL canDo)
    
        # More Accessor Functions
        void			SetImageStartTime()
        #void			SetImageStartTime(time_t startTime)
        void			SetImageStartTime(int mon, int dd, int yy, int hr, int min, int sec)
        #void			SetImageStartTime(struct tm *pStartTime)
        void			SetSubFrame(int nLeft, int nTop)
        void			GetSubFrame(int &nLeft, int &nTop)
        void			SetBinning(unsigned short nHoriz, unsigned short nVert)
        void			GetBinning(unsigned short &nHoriz, unsigned short &nVert)
        void			AddHistory(string str)
    
        # File IO Routines
        #SBIG_FILE_ERROR SaveImage(const char *pFullPath, SBIG_IMAGE_FORMAT fmt = SBIF_DEFAULT)
        #SBIG_FILE_ERROR OpenImage(const char *pFullPath)
    
        # Utility Functions
        MY_LOGICAL		AllocateImageBuffer(int height, int width)
        void			CreateSBIGHeader(char *pHeader, MY_LOGICAL isCompressed)
        MY_LOGICAL		ParseHeader(char *pHeader, MY_LOGICAL &isCompressed)
        #SBIG_FILE_ERROR SaveCompressedImage(const char *pFullPath, char *pHeader)
        #SBIG_FILE_ERROR ReadCompressedImage(FILE *fh)
        #SBIG_FILE_ERROR SaveUncompressedImage(const char *pFullPath, char *pHeader)
        #SBIG_FILE_ERROR ReadUncompressedImage(FILE *fh)
        int				CompressSBIGData(unsigned char *pCmpData, int imgRow)
        void			IntelCopyBytes(unsigned char *pRevData, int imgRow)
        void			AutoBackgroundAndRange()
        #string			GetFileErrorString(SBIG_FILE_ERROR err)
        unsigned short	GetAveragePixelValue()
        unsigned short	GetAveragePixelValue(int left, int top, int width, int height)
        void			GetFormattedImageInfo(string &iiStr, MY_LOGICAL htmlFormat)
    
        # Image Processing Funcions
        void			VerticalFlip()
        void			HorizontalFlip()
        #SBIG_FILE_ERROR DarkSubtract(CSBIGImg *pImg)
        #SBIG_FILE_ERROR FlatField(CSBIGImg *pImg)
        
        # Color Image Processing
        void			RemoveBayerColor()


cdef extern from "../src/csbigcam.h":
    cdef enum SBIG_DEVICE_TYPE:
        DEV_USB = 0x7F00

    ctypedef int PAR_ERROR

    ctypedef enum SBIG_DARK_FRAME:
        SBDF_LIGHT_ONLY, SBDF_DARK_ONLY, SBDF_DARK_ALSO

    ctypedef enum SHUTTER_COMMAND:
        SC_LEAVE_SHUTTER, SC_OPEN_SHUTTER, SC_CLOSE_SHUTTER, SC_INITIALIZE_SHUTTER, \
            SC_OPEN_EXT_SHUTTER, SC_CLOSE_EXT_SHUTTER

    ctypedef enum CFW_MODEL_SELECT:
        CFWSEL_UNKNOWN, CFWSEL_CFW2, CFWSEL_CFW5, CFWSEL_CFW8, CFWSEL_CFWL, CFWSEL_CFW402, CFWSEL_AUTO, CFWSEL_CFW6A, \
        CFWSEL_CFW10, CFWSEL_CFW10_SERIAL, CFWSEL_CFW9, CFWSEL_CFWL8, CFWSEL_CFWL8G, CFWSEL_CFW1603, \
        CFWSEL_FW5_STX, CFWSEL_FW5_8300, CFWSEL_FW8_8300, CFWSEL_FW7_STX, CFWSEL_FW8_STT

    ctypedef enum CFW_POSITION:
        CFWP_UNKNOWN, CFWP_1, CFWP_2, CFWP_3, CFWP_4, CFWP_5, CFWP_6, CFWP_7, CFWP_8, CFWP_9, CFWP_10

    ctypedef  int CFW_ERROR

    ctypedef enum CFW_COM_PORT:
        CFWPORT_COM1=1, CFWPORT_COM2, CFWPORT_COM3, CFWPORT_COM4

    ctypedef enum CFW_STATUS:
        CFWS_UNKNOWN, CFWS_IDLE, CFWS_BUSY

    ctypedef enum CCD_REQUEST:
        CCD_IMAGING, CCD_TRACKING, CCD_EXT_TRACKING

    cdef cppclass CSBIGCam:
        CSBIGCam(int type)

        void Init()

        # Error Reporting Routines
        PAR_ERROR 	GetError()
        #string 		GetErrorString()
        string 		GetErrorString(PAR_ERROR err)
        #PAR_COMMAND GetCommand()

        # active
        CCD_REQUEST GetActiveCCD()
        void SetActiveCCD(CCD_REQUEST)

        # Setters /Getters
        unsigned short GetFirmwareVersion()
        double GetExposureTime()
        void SetExposureTime(double exp)
        unsigned short GetReadoutMode()
        void SetReadoutMode(unsigned short rm)
        # CAMERA_TYPE GetCameraType()
        # ABG_STATE7 GetABGState()
        # void SetABGState(ABG_STATE7 abgState)
        void SetSubFrame(int nLeft,  int nTop,  int nWidth,  int nHeight)
        void GetSubFrame(int &nLeft, int &nTop, int &nWidth, int &nHeight)
        PAR_ERROR GetReadoutInfo(double &pixelWidth, double &pixelHeight, double &eGain)

        # Driver/Device Routines
        PAR_ERROR OpenDriver()
        PAR_ERROR CloseDriver()
        #PAR_ERROR OpenDevice(OpenDeviceParams odp)
        PAR_ERROR CloseDevice()
        #PAR_ERROR GetDriverInfo(DRIVER_REQUEST request, GetDriverInfoResults0 &gdir)

        # High-Level Exposure Related Commands
        PAR_ERROR GrabSetup(CSBIGImg *pImg, SBIG_DARK_FRAME dark)
        PAR_ERROR GrabMain (CSBIGImg *pImg, SBIG_DARK_FRAME dark)
        PAR_ERROR GrabImage(CSBIGImg *pImg, SBIG_DARK_FRAME dark)
        #void 	  GetGrabState(GRAB_STATE &grabState, double &percentComplete)
        PAR_ERROR Readout  (CSBIGImg *pImg, SBIG_DARK_FRAME dark) nogil

        # Low-Level Exposure Related Commands
        PAR_ERROR StartExposure(SHUTTER_COMMAND shutterState)
        PAR_ERROR EndExposure()
        PAR_ERROR IsExposureComplete(MY_LOGICAL &complete)
        #PAR_ERROR StartReadout(StartReadoutParams srp)
        PAR_ERROR EndReadout()
        #PAR_ERROR ReadoutLine(ReadoutLineParams rlp, MY_LOGICAL darkSubtract, unsigned short *dest)
        PAR_ERROR DumpLines(unsigned short noLines)

        # Temperature Related Commands
        PAR_ERROR GetCCDTemperature(double &ccdTemp)
        PAR_ERROR SetTemperatureRegulation(MY_LOGICAL enable, double setpoint)
        PAR_ERROR QueryTemperatureStatus(MY_LOGICAL &enabled, double &ccdTemp, double &setpointTemp, double &percentTE)

        # Control Related Commands
        #PAR_ERROR ActivateRelay(CAMERA_RELAY relay, double time)
        #PAR_ERROR IsRelayActive(CAMERA_RELAY relay, MY_LOGICAL &active)
        #PAR_ERROR AOTipTilt(AOTipTiltParams attp)
        #PAR_ERROR CFWCommand(CFWParams cfwp, CFWResults &cfwr)
        PAR_ERROR InitializeShutter()

        # General Purpose Commands
        PAR_ERROR EstablishLink()
        string 	  GetCameraTypeString()
        PAR_ERROR GetFullFrame(int &nWidth, int &nHeight)
        PAR_ERROR GetFormattedCameraInfo(string &ciStr, MY_LOGICAL htmlFormat)

        # Utility functions
        MY_LOGICAL 		CheckLink()
        unsigned short 	DegreesCToAD(double degC, MY_LOGICAL ccd)
        double 			ADToDegreesC(unsigned short ad, MY_LOGICAL ccd)

        # CFW Functions
        CFW_MODEL_SELECT	GetCFWModel()
        PAR_ERROR 			SetCFWModel(CFW_MODEL_SELECT cfwModel, CFW_COM_PORT comPort)
        PAR_ERROR 			SetCFWPosition(CFW_POSITION position)
        PAR_ERROR 			GetCFWPositionAndStatus(CFW_POSITION &position, CFW_STATUS &status)
        PAR_ERROR 			GetCFWMaxPosition(CFW_POSITION &position)
        CFW_ERROR 			GetCFWError()
        string 				GetCFWErrorString(CFW_ERROR err)
        #string 				GetCFWErrorString()

        # Allows access directly to driver
        PAR_ERROR SBIGUnivDrvCommand(short command, void *Params, void *Results)



import logging
from typing import Any, Optional

from pyobs.object import Object

from .sbigudrv import *


log = logging.getLogger(__name__)


class SbigDriver(Object):
    """The underlying driver for the pyobs SBIG modules."""

    def __init__(self, filter_wheel: str = 'UNKNOWN', **kwargs: Any):
        """Initializes a new SbigCamera.

        Args:
            filter_wheel: Name of filter wheel, if any.
        """
        Object.__init__(self, **kwargs)

        # create cam
        self.camera = SBIGCam()

        # filter wheel
        self.filter_wheel = FilterWheelModel[filter_wheel]

        # active lock
        self._lock_active = threading.Lock()

    def open(self) -> None:
        """Open module.

        Raises:
            ValueError: If cannot connect to camera or set filter wheel.
        """

        # set filter wheel model
        if self.filter_wheel != FilterWheelModel.UNKNOWN:
            log.info('Initialising filter wheel...')
            try:
                self.camera.set_filter_wheel(self.filter_wheel)
            except ValueError as e:
                raise ValueError('Could not set filter wheel: %s' % str(e))

        # open driver
        log.info('Opening SBIG driver...')
        try:
            self.camera.establish_link()
        except ValueError as e:
            raise ValueError('Could not establish link: %s' % str(e))

    def full_frame(self, sensor: ActiveSensor) -> Tuple[int, int, int, int]:
        """Return full frame."""
        with self._lock_active:
            # TODO: maybe rethink this: should the camera return the full frame for the current binning and
            # not for 1x1?
            self.camera.binning = (1, 1)
            self.camera.sensor = sensor
            return self.camera.full_frame

    def start_exposure(self, sensor: ActiveSensor, img: SBIGImg, shutter: bool, exposure_time: float,
                       window: Optional[Tuple[int, int, int, int]] = None,
                       binning: Optional[Tuple[int, int]] = None) -> None:
        """Start an exposure.

        Args:
            sensor: Sensor to use for exposure.
            img: Image to write into.
            shutter: Whether to open shutter.
            exposure_time: Exposure time in secs.
            window: Window to use.
            binning: Binning to use.
        """

        # do all settings within a mutex
        with self._lock_active:
            # set active sensor
            self.camera.sensor = sensor

            # set exposure time, window and binnint
            self.camera.exposure_time = exposure_time
            self.camera.window = window
            self.camera.binning = binning

            # start exposure
            self.camera.start_exposure(img, shutter)

    def has_exposure_finished(self, sensor: ActiveSensor) -> bool:
        """Whether an exposure has finished

        Args:
            sensor: Sensor to use for exposure.

        Returns:
            Exposure finished or not.
        """
        with self._lock_active:
            self.camera.sensor = sensor
            return self.camera.has_exposure_finished()

    def end_exposure(self, sensor: ActiveSensor) -> None:
        """End an exposure."""
        with self._lock_active:
            self.camera.sensor = sensor
            return self.camera.end_exposure()

    def readout(self, sensor: ActiveSensor, img: SBIGImg, shutter: bool) -> None:
        """Readout image.

        Args:
            sensor: Camera to use for exposure.
            img: Image to read into.
            shutter: Whether shutter was open.
        """
        with self._lock_active:
            self.camera.sensor = sensor
            return self.camera.readout(img, shutter)


__all__ = ['SbigDriver']
